/*
 * distromatic - tool for RPM based repositories
 *
 * Copyright (C) 2013-2020 by Silvan Calarco <silvan.calarco@mambasoft.it>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sqlite3.h>
#include "headerlist.h"
#include "changelog.h"
#include "backend-sqlite3.h"
#include "functions.h"

char sqlite3_query[PATH_MAX];
sqlite3_stmt *stmt;

#define sqlite3_transaction_maxsize 10485760
long sqlite3_transaction_size;

int SQLite_init_table(sqlite3 *db, const char* table_name, const char* table_query) {

    char myquery[PATH_MAX];

    snprintf(myquery, PATH_MAX, "CREATE TABLE IF NOT EXISTS %s(%s)", table_name, table_query);

    if (sqlite3_exec(db, myquery, 0, 0, 0)) {
       fprintf(stderr, "ERROR: SQLite: (%s) %s\n", myquery, sqlite3_errmsg(db));
       return 2;
    }
    return 0;
}

int SQLite_begin_transaction(sqlite3 *db) {

    if (sqlite3_exec(db, "BEGIN TRANSACTION;", 0, 0, 0)) {
       fprintf(stderr, "ERROR: SQLite: (BEGIN TRANSACTION) %s\n", sqlite3_errmsg(db));
       return 1;
    }

    sqlite3_transaction_size = 0;
    return 0;
}

int SQLite_commit_transaction(sqlite3 *db) {
   if (sqlite3_exec(db, "COMMIT;", 0, 0, 0)) {
       fprintf(stderr, "ERROR: SQLite: (COMMIT) %s\n", sqlite3_errmsg(db));
       return 1;
   }
   sqlite3_transaction_size = 0;
   return 0;
}

int SQLite_print_contents_subtree(sqlite3 *db,
                            struct fileTree* ft,
			    struct configTag* ct,
			    int arch) {

   int thisrep, i;

   while (ft) {
      i = 0;
      thisrep = 0;
      for (i = 0; i < ft->numproviders; i++) {
         if (ft->provider[i]->altrepository == ct->repository_level) {
            thisrep = 1;
            break;
         }
      }
      if (thisrep) {
         snprintf(sqlite3_query, PATH_MAX, "INSERT INTO files VALUES(%ld,?,%ld,%ld,%ld,%d);",
            ft->id,
            (ft->firstchild?ft->firstchild->id:-1),
            (ft->next?ft->next->id:-1),
            (ft->parent?ft->parent->id:-1),
            ft->numproviders);
         sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL);
         sqlite3_bind_text(stmt, 1, ft->name, -1, SQLITE_STATIC);
         if (sqlite3_step(stmt) != SQLITE_DONE) {
            fprintf(stderr, "ERROR: sqlite3_step: %s (%s)\n", sqlite3_errmsg(db), sqlite3_query);
            return 1;
         }
         sqlite3_finalize(stmt);

         sqlite3_transaction_size += strlen(sqlite3_query);
         if (sqlite3_transaction_size >= sqlite3_transaction_maxsize) {
            SQLite_commit_transaction(db);
            SQLite_begin_transaction(db);
         }

         if (ft->firstchild) {
            if (SQLite_print_contents_subtree(db, ft->firstchild, ct, arch)) return 1;
         }
      }
      ft=ft->next;
   }
   return 0;
}

#define SQLITE_TABLE_files "id INTEGER PRIMARY KEY, "\
   "name STRING, firstchild INTEGER, next INTEGER, parent INTEGER, numproviders INTEGER"
/*     struct headerList **provider; */

#define SQLITE_TABLE_packages_files_rel "id INTEGER PRIMARY KEY, id_package INTEGER, id_file INTEGER, "\
   "id_user INTEGER, id_group INTEGER, flags INTEGER, "\
   "name STRING, version STRING, release STRING"
#define SQLITE_TABLE_fileusers "id INTEGER PRIMARY KEY, name STRING"
#define SQLITE_TABLE_filegroups "id INTEGER PRIMARY KEY, name STRING"

int generateSQLite_files(struct configTag* ct, sqlite3 *db, int arch) {

   int i;
   struct headerList* currpackage;
   struct fileUserList *fileUser = ct->fileuserlist[arch];
   struct fileGroupList *fileGroup = ct->filegrouplist[arch];

   snprintf(sqlite3_query, PATH_MAX, "packages_files_rel");
   SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_packages_files_rel);

   snprintf(sqlite3_query, PATH_MAX, "files");
   SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_files);

   snprintf(sqlite3_query, PATH_MAX, "fileusers");
   SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_fileusers);

   snprintf(sqlite3_query, PATH_MAX, "filegroups");
   SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_filegroups);

   while (fileUser) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO fileusers VALUES(%ld,?);",
               fileUser->id);
      if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
             fprintf(stderr, "ERROR: sqlite3_prepare_v2: %s (%s)\n", sqlite3_errmsg(db), sqlite3_query);
             return 1;
      }
      sqlite3_bind_text(stmt, 1, fileUser->name, -1, SQLITE_STATIC);
      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "WARNING: SQLite: (%s %s) %s; ignoring\n", sqlite3_query,
                 fileUser->name, sqlite3_errmsg(db));
      }
      sqlite3_finalize(stmt);
      fileUser = fileUser->next;
   }

   while (fileGroup) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO filegroups VALUES(%ld,?);",
               fileGroup->id);
      if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
         fprintf(stderr, "ERROR: sqlite3_prepare_v2: %s (%s)\n", sqlite3_errmsg(db), sqlite3_query);
         return 1;
      }
      sqlite3_bind_text(stmt, 1, fileGroup->name, -1, SQLITE_STATIC);
      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "WARNING: SQLite: (%s %s) %s; ignoring\n", sqlite3_query,
                 fileGroup->name, sqlite3_errmsg(db));
      }
      sqlite3_finalize(stmt);
      fileGroup = fileGroup->next;
   }

   SQLite_begin_transaction(db);
   if (SQLite_print_contents_subtree(db, ct->filetree[arch], ct, arch)) return 1;

   currpackage = ct->headerlist[arch];
   while (currpackage) {
      if (currpackage->altrepository == ct->repository_level) {
         /* packages <-> files relations */
         for (i = 0; i < currpackage->filenamecount; i++) {
            snprintf(sqlite3_query, PATH_MAX, "INSERT INTO packages_files_rel VALUES(NULL,%ld,%ld,%ld,%ld,%d,?,?,?);",
               currpackage->id,
               currpackage->file[i]->id,
               currpackage->fileuser[i]->id,
               currpackage->filegroup[i]->id,
               currpackage->fileflags[i]);
            if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
               fprintf(stderr, "ERROR: sqlite3_prepare_v2: %s (%s)\n", sqlite3_errmsg(db), sqlite3_query);
               return 1;
            }
            sqlite3_bind_text(stmt, 1, currpackage->name, -1, SQLITE_STATIC);
            sqlite3_bind_text(stmt, 2, currpackage->version, -1, SQLITE_STATIC);
            sqlite3_bind_text(stmt, 3, currpackage->release, -1, SQLITE_STATIC);
            if (sqlite3_step(stmt) != SQLITE_DONE) {
               fprintf(stderr, "ERROR: sqlite3_step: %s (%s)\n", sqlite3_query, sqlite3_errmsg(db));
               return 3;
            }
            sqlite3_finalize(stmt);
            sqlite3_transaction_size += strlen(sqlite3_query);
            if (sqlite3_transaction_size >= sqlite3_transaction_maxsize) {
               SQLite_commit_transaction(db);
               SQLite_begin_transaction(db);
            }
         }
      }
      currpackage = currpackage->next;
   }
   SQLite_commit_transaction(db);
   return 0;
}

#define SQLITE_TABLE_provided "id INTEGER PRIMARY KEY, "\
   "name STRING, flags INTEGER, numproviders INTEGER"

int generateSQLite_provided(struct configTag* ct, sqlite3 *db, int arch) {

   struct providedList* provided;
   int i, thisrep;

   SQLite_init_table(db, "provided", SQLITE_TABLE_provided);
   SQLite_begin_transaction(db);

   provided = ct->providedlist_idx[arch][0];
   while (provided) {
      i = 0;
      thisrep = 0;
      for (i = 0; i < provided->numproviders; i++) {
         if (provided->provider[i]->altrepository == ct->repository_level) {
            thisrep = 1;
            break;
         }
      }
      if (thisrep) {
        snprintf(sqlite3_query, PATH_MAX, "INSERT INTO provided VALUES(%ld,?,%ld,%d);",
            provided->id,
            provided->flags,
            provided->numproviders);
        if (sqlite3_prepare_v3(db, sqlite3_query, -1, 0, &stmt, NULL)) {
            fprintf(stderr, "ERROR: sqlite3_prepare_v2: %s (%s)\n",
                sqlite3_errmsg(db), sqlite3_query);
           return 1;
        }
        sqlite3_bind_text(stmt, 1, provided->name, -1, SQLITE_STATIC);
        if (sqlite3_step(stmt) != SQLITE_DONE) {
            fprintf(stderr, "WARNING: SQLite: (%s %s) %s; ignoring\n", sqlite3_query,
                provided->name, sqlite3_errmsg(db));
        }
        sqlite3_finalize(stmt);
      }
      provided = provided->next;
   }
   SQLite_commit_transaction(db);
   return 0;
}

#define SQLITE_TABLE_packagers "id INTEGER PRIMARY KEY, "\
   "name STRING, role INTEGER, changes_count INTEGER, packages_count INTEGER"
/*     char *alias[PACKAGER_MAXALIASES]; */

int generateSQLite_packagers(sqlite3 *db) {

   struct Packager* packager = firstPackager();

   SQLite_init_table(db, "packagers", SQLITE_TABLE_packagers);
   SQLite_begin_transaction(db);

   while (packager) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO packagers VALUES(NULL,?,%d,%d,%d);",
         packager->role,
         packager->changes_count,
         packager->packages_count);

      if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
      }
      sqlite3_bind_text(stmt, 1, packager->name, -1, SQLITE_STATIC);
      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
         return 3;
      }
      packager->id = sqlite3_last_insert_rowid(db);
      sqlite3_finalize(stmt);
      packager = packager->next;
   }
   SQLite_commit_transaction(db);
   return 0;
}

long generateSQLite_add_changelog(sqlite3 *db, struct changeLog* firstchangelog, long id_source) {

   struct changeLog* changelog = firstchangelog;
   long lastchangelogid = 0;

   while (changelog) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO changelog VALUES(NULL,%ld,%ld,%ld,?,?);",
         id_source,
         changelog->time,
         changelog->pkg->id);

      if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
      }
      sqlite3_bind_text(stmt, 1, changelog->release, -1, SQLITE_STATIC);
      sqlite3_bind_text(stmt, 2, changelog->text, -1, SQLITE_STATIC);
      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
         return 3;
      }
      changelog->id = sqlite3_last_insert_rowid(db);
      lastchangelogid = changelog->id;
      sqlite3_finalize(stmt);
      changelog = changelog->next;
   }
   return lastchangelogid;
}

#define SQLITE_TABLE_packages "id INTEGER PRIMARY KEY, "\
   "name STRING, altrepository INTEGER, epoch INTEGER, version STRING, release STRING, summary STRING, "\
   "groupdescr STRING, description STRING, size INTEGER, id_source INTEGER"

#define SQLITE_TABLE_obsoletes "id INTEGER PRIMARY KEY, "\
   "id_package INTEGER, obsoletename STRING, obsoleteflags INTEGER, obsoleteversion STRING"

#define SQLITE_TABLE_provides "id INTEGER PRIMARY KEY, "\
   "id_package INTEGER, id_provided INTEGER, provideflags INTEGER, provideversion STRING"

#define SQLITE_TABLE_requires "id INTEGER PRIMARY KEY, "\
   "id_package INTEGER, id_provided STRING, requireflags INTEGER, requireversion STRING"

int
generateSQLite_packages(struct configTag *ct, sqlite3 *db, int arch) {

   struct headerList* currpackage;
   int i;

   SQLite_init_table(db, "packages", SQLITE_TABLE_packages);
   SQLite_init_table(db, "obsoletes", SQLITE_TABLE_obsoletes);
   SQLite_init_table(db, "provides", SQLITE_TABLE_provides);
   SQLite_init_table(db, "requires", SQLITE_TABLE_requires);
   SQLite_begin_transaction(db);

   currpackage = ct->headerlist[arch];
   while (currpackage) {
      if (currpackage->altrepository == ct->repository_level) {
         snprintf(sqlite3_query, PATH_MAX, "INSERT INTO packages VALUES(NULL,?,%d,%ld,?,?,?,?,?,%ld,%ld);",
            currpackage->altrepository,
            currpackage->epoch,
            currpackage->size,
            currpackage->sourceheader->id);
         if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
            fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
            return 1;
         }

         sqlite3_bind_text(stmt, 1, currpackage->name, -1, SQLITE_STATIC);
         sqlite3_bind_text(stmt, 2, currpackage->version, -1, SQLITE_STATIC);
         sqlite3_bind_text(stmt, 3, currpackage->release, -1, SQLITE_STATIC);
         sqlite3_bind_text(stmt, 4, currpackage->summary, -1, SQLITE_STATIC);
         sqlite3_bind_text(stmt, 5, currpackage->group, -1, SQLITE_STATIC);
         sqlite3_bind_text(stmt, 6, currpackage->description, -1, SQLITE_STATIC);

         if (sqlite3_step(stmt) != SQLITE_DONE) {
            fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
            return 3;
         }
         currpackage->id = sqlite3_last_insert_rowid(db);
         sqlite3_finalize(stmt);

         /* obsoletes */
         for (i = 0; i < currpackage->obsoletecount; i++) {
            snprintf(sqlite3_query, PATH_MAX, "INSERT INTO obsoletes VALUES(NULL,%ld,?,%d,?);",
               currpackage->id,
               currpackage->obsoleteflags[i]);

            if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
               fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
            }
            sqlite3_bind_text(stmt, 1, currpackage->obsoletename[i], -1, SQLITE_STATIC);
            sqlite3_bind_text(stmt, 2, currpackage->obsoleteversion[i], -1, SQLITE_STATIC);
            if (sqlite3_step(stmt) != SQLITE_DONE) {
               fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
               return 3;
            }
            sqlite3_finalize(stmt);
         }

         /* provides */
         for (i = 0; i < currpackage->providecount; i++) {
            snprintf(sqlite3_query, PATH_MAX, "INSERT INTO provides VALUES(NULL,%ld,%ld,%d,?);",
               currpackage->id,
               currpackage->provided[i]->id,
               currpackage->provideflags[i]);

            if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
               fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
            }
            sqlite3_bind_text(stmt, 1, currpackage->provideversion[i], -1, SQLITE_STATIC);
            if (sqlite3_step(stmt) != SQLITE_DONE) {
               fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
               return 3;
            }
            sqlite3_finalize(stmt);
         }

         /* requires */
         for (i = 0; i < currpackage->requirecount; i++) {
            if (currpackage->require[i]->resolved) {
               snprintf(sqlite3_query, PATH_MAX, "INSERT INTO requires VALUES(NULL,%ld,%ld,%ld,?);",
                  currpackage->id,
                  currpackage->require[i]->resolved->id,
                  currpackage->require[i]->flags);

               if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
                  fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
               }
               sqlite3_bind_text(stmt, 1, currpackage->require[i]->version, -1, SQLITE_STATIC);
               if (sqlite3_step(stmt) != SQLITE_DONE) {
                  fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                  return 3;
               }
               sqlite3_finalize(stmt);
            } else if (strstr(currpackage->require[i]->name,"rpmlib(") != currpackage->require[i]->name) {
               fprintf(stderr, "WARNING: package %s require %s not in provided list\n", currpackage->name, currpackage->require[i]->name);
            }
         }
      }
      currpackage = currpackage->next;
   }
   SQLite_commit_transaction(db);
   return 0;
}

#define SQLITE_TABLE_sources_source "id INTEGER PRIMARY KEY, id_source INTEGER, source STRING"
#define SQLITE_TABLE_sources_patch "id INTEGER PRIMARY KEY, id_source INTEGER, patch STRING"
#define SQLITE_TABLE_sources "id INTEGER PRIMARY KEY, "\
   "name STRING, altrepository INTEGER, epoch INTEGER, version STRING, release STRING, summary STRING, id_packager INTEGER, "\
   "groupdescr STRING, description STRING, url STRING, license STRING, arch STRING, buildarchs STRING, " \
   "excludearch STRING, buildtime INTEGER, size INTEGER, id_changelog INTEGER"
//#define SQLITE_TABLE_sources_files_rel "id INTEGER PRIMARY KEY, id_source INTEGER, id_files INTEGER"
//#define SQLITE_TABLE_sources_provided_rel "id INTEGER PRIMARY KEY, id_source INTEGER, id_provided INTEGER"
#define SQLITE_TABLE_changelog "id INTEGER PRIMARY KEY, "\
   "id_source INTEGER, time INTEGER, id_packager INTEGER, release STRING, text STRING"
#define SQLITE_TABLE_buildrequires "id INTEGER PRIMARY KEY, "\
   "id_source INTEGER, buildrequirename STRING, buildrequireflags INTEGER, buildrequireversion STRING"

int
generateSQLite_sources(struct configTag *ct, sqlite3 *db) {

    struct headerSourceList* currsource;
    int i;
    long nextchangelogid = 1;

    SQLite_init_table(db, "sources", SQLITE_TABLE_sources);
    SQLite_init_table(db, "sources_source", SQLITE_TABLE_sources_source);
    SQLite_init_table(db, "sources_patch", SQLITE_TABLE_sources_patch);
//    SQLite_init_table(db, "sources_files_rel", SQLITE_TABLE_sources_files_rel);
//    SQLite_init_table(db, "sources_provided_rel", SQLITE_TABLE_sources_provided_rel);
    SQLite_init_table(db, "changelog", SQLITE_TABLE_changelog);
    SQLite_init_table(db, "buildrequires", SQLITE_TABLE_buildrequires);
    SQLite_begin_transaction(db);

    currsource = ct->headersourcelist;
    while (currsource != NULL) {
       if (currsource->altrepository == ct->repository_level) {
          snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources VALUES(%ld,?,%d,%ld,?,?,?,%ld,?,?,?,?,?,?,?,%ld,%ld,%ld);",
             currsource->id,
             currsource->altrepository,
             currsource->epoch,
             currsource->packager->id,
             currsource->buildtime,
             currsource->size,
             (currsource->changelog?nextchangelogid:0));
          if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
          }

          sqlite3_bind_text(stmt, 1, currsource->name, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 2, currsource->version, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 3, currsource->release, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 4, currsource->summary, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 5, currsource->group, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 6, currsource->description, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 7, currsource->url, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 8, currsource->license, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 9, currsource->arch, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 10, currsource->buildarchs, -1, SQLITE_STATIC);
          sqlite3_bind_text(stmt, 11, currsource->excludearch, -1, SQLITE_STATIC);

          if (sqlite3_step(stmt) != SQLITE_DONE) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
             return 3;
          }
          currsource->id = sqlite3_last_insert_rowid(db);
          sqlite3_finalize(stmt);

          /* source */
          if (currsource->source) {
             i=0;
             while (currsource->source[i]) {
                snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources_source VALUES(NULL,%ld,?);",
                   currsource->id);

                if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
                   fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                }
                sqlite3_bind_text(stmt, 1, currsource->source[i], -1, SQLITE_STATIC);
                if (sqlite3_step(stmt) != SQLITE_DONE) {
                      fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                   return 3;
                }
                sqlite3_finalize(stmt);
                i++;
             }
          }

          /* patch */
          if (currsource->patch) {
             i=0;
             while (currsource->patch[i]) {
                snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources_patch VALUES(NULL,%ld,?);",
                   currsource->id);

                if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
                   fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                }
                sqlite3_bind_text(stmt, 1, currsource->patch[i], -1, SQLITE_STATIC);
                if (sqlite3_step(stmt) != SQLITE_DONE) {
                   fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                   return 3;
                }
                sqlite3_finalize(stmt);
                i++;
             }
          }

          /* buildrequires */
          for (i = 0; i < currsource->requirecount; i++) {
             snprintf(sqlite3_query, PATH_MAX, "INSERT INTO buildrequires VALUES(NULL,%ld,?,%ld,?);",
                currsource->id,
                currsource->require[i]->flags);

             if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
                fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
             }
             sqlite3_bind_text(stmt, 1, currsource->require[i]->name, -1, SQLITE_STATIC);
             sqlite3_bind_text(stmt, 2, currsource->require[i]->version, -1, SQLITE_STATIC);
             if (sqlite3_step(stmt) != SQLITE_DONE) {
                fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                return 3;
             }
             sqlite3_finalize(stmt);
          }

       /* sources <-> provided relations (buildrequirements) */
/*       for (i=0; i<currsource->providecount; i++) {
          snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources_provided_rel VALUES(NULL,%d,%d);",
             currsource->id,
             currsource->required[i]->id);

          if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
          }
          if (sqlite3_step(stmt) != SQLITE_DONE) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
             return 3;
          }
          sqlite3_finalize(stmt);
       }*/

          if (currsource->changelog) {
             nextchangelogid = generateSQLite_add_changelog(db, currsource->changelog, currsource->id) + 1;
          }
       }
       currsource = currsource->next;
    }
    SQLite_commit_transaction(db);
    return 0;
}

int
generateSQLiteFiles(struct configTag *ct)
{
    sqlite3 *db = NULL;
    char dbname[PATH_MAX];
    int i;

    for (i = 0; i < ARCHS_MAX && ct->arch[i]; i++) {
       snprintf(dbname, PATH_MAX, "%s%s-%s-files.db", ct->repository_dir, ct->tag, ct->arch[i]);
       unlink(dbname);

       if (sqlite3_open_v2(dbname, &db, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL)) {
          logmsg(LOG_ERROR, "sqlite3_open_v2: %s (%s)", dbname, sqlite3_errmsg(db));
          if (db) sqlite3_close(db);
          return 1;
       }
       if (generateSQLite_files(ct, db, i)) return 1;
       if (db) sqlite3_close(db);
    }
    return 0;
}

int
generateSQLite(struct configTag *ct)
{
    sqlite3 *db = NULL;
    char dbname[PATH_MAX];
    int i;

    for (i = 0; i < ARCHS_MAX && ct->arch[i]; i++) {
       snprintf(dbname, PATH_MAX, "%s%s-%s.db", ct->repository_dir, ct->tag, ct->arch[i]);
       unlink(dbname);

       if (sqlite3_open_v2(dbname, &db, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL)) {
          logmsg(LOG_ERROR, "sqlite3_open_v2: %s %s", sqlite3_errmsg(db), dbname);
          if (db) sqlite3_close(db);
          return 1;
       }

       if (generateSQLite_packages(ct, db, i)) return 1;
       if (generateSQLite_provided(ct, db, i)) return 1;
       sqlite3_close(db);
    }

    snprintf(dbname, PATH_MAX, "%s%s-sources.db", ct->repository_dir, ct->tag);
    unlink(dbname);

    if (sqlite3_open_v2(dbname, &db, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL)) {
       if (db) sqlite3_close(db);
       return 1;
    }

    generateSQLite_sources(ct, db);
    generateSQLite_packagers(db);

    sqlite3_close(db);
    return 0;
}
