/*
 * distromatic - tool for RPM based repositories
 *
 * Copyright (C) 2004-2021 by Silvan Calarco <silvan.calarco@mambasoft.it>
 * Copyright (C) 2006 by Davide Madrisan <davide.madrisan@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#if STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>
#include <libiberty/libiberty.h>

#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#if HAVE_STRINGS_H
# include <strings.h>
#endif

#ifdef HAVE_LIBIBERTY_H
# include <libiberty.h>
#endif

/* Tell glibc's <time.h> to provide a prototype for strptime() */
#ifndef __USE_XOPEN
# define __USE_XOPEN
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include "distromatic.h"
#include "config.h"
#include "functions.h"
#include "changelog.h"

#define UNW_LOCAL_ONLY
#include <elfutils/libdwfl.h>
#include <libunwind.h>
#include <assert.h>

static int debug_log = 0;
static struct configDefaults configdefaults;
static struct configTag *firstconfigtag = NULL;

/* Obtain a backtrace and print it to stdout. */
static void debugInfo(FILE *out, const void* ip)
{
	char *debuginfo_path=NULL;

	Dwfl_Callbacks callbacks={
		.find_elf=dwfl_linux_proc_find_elf,
		.find_debuginfo=dwfl_standard_find_debuginfo,
		.debuginfo_path=&debuginfo_path,
	};

	Dwfl* dwfl=dwfl_begin(&callbacks);
	assert(dwfl!=NULL);

	assert(dwfl_linux_proc_report (dwfl, getpid())==0);
	assert(dwfl_report_end (dwfl, NULL, NULL)==0);

	Dwarf_Addr addr = (uintptr_t)ip;

	Dwfl_Module* module=dwfl_addrmodule (dwfl, addr);

	const char* function_name = dwfl_module_addrname(module, addr);

	fprintf(out, "%s(",function_name);

	Dwfl_Line *line=dwfl_getsrc (dwfl, addr);
	if(line!=NULL)
	{
		int nline;
		Dwarf_Addr addr;
		const char* filename=dwfl_lineinfo (line, &addr,&nline,NULL,NULL,NULL);
		fprintf(out, "%s:%d",strrchr(filename,'/')+1,nline);
	}
	else
	{
    const char *module_name = dwfl_module_info(module,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    fprintf(out, "in %s", module_name);
	}
}

void backtraceHandler(int sig)
{
  unw_context_t uc;
  unw_getcontext(&uc);

  unw_cursor_t cursor;
  unw_init_local(&cursor, &uc);

  fprintf(stderr, "*** Received signal %d; stack trace: ***\n", sig);

  while(unw_step(&cursor)>0) {
    unw_word_t ip;
    unw_get_reg(&cursor, UNW_REG_IP, &ip);

    unw_word_t offset;
    char name[32];
    assert(unw_get_proc_name(&cursor, name,sizeof(name), &offset)==0);

    fprintf(stderr, "\tat ");
    debugInfo(stderr, (void*)(ip-4));
    fprintf(stderr, ")\n");

    if(strcmp(name,"main")==0)
      break;
  }
   exit(1);
}

struct configTag* findRepositoryByTag(const char *tag)
{
   struct configTag *ct = firstconfigtag;

   while ((ct) && (strncmp(ct->tag,tag,PATH_MAX))) ct = ct->next;

   if (ct) return ct; else return NULL;
}

struct configTag* read_configuration(const char *confFile)
{
     FILE *fin;

     char *vartok, *valuetok;
     char input[PATH_MAX];
     char buf[PATH_MAX];
     int i, curraltrep = 0;

     struct configTag *newconfigtag, *currconfigtag = NULL;
     struct Packager *currmaintainer;

     if ((fin = fopen(confFile, "r")) == NULL) {
          return NULL;
     }

     unsigned int configsection = 0;
     configdefaults.html_basedir = NULL;

     configdefaults.arch[0] = DEFAULT_ARCH;
     for (i = 1; i<ARCHS_MAX; i++)
	configdefaults.arch[i] = NULL;

     while (fgets(input, PATH_MAX, fin)) {

          if (input[0] == '#') {
               continue;
          }

          vartok = (char *) strtok(input, "=\n");
          if (!vartok) {        /* skip blank lines */
               continue;
          }
          strip_separators(vartok, " \t\n");

	  valuetok = (char *) strtok(NULL, "\n");
          strip_separators(valuetok, " \t\n");

          if ((vartok[0]=='[') && (vartok[strlen(vartok)-1]==']')) {
               /* found a `[defaults]' tag */
               if (!strncmp(&vartok[1], "defaults", strlen(vartok)-2)) {
                    configsection = CONF_DEFAULTS_SECTION;
               } else if (!strncmp(&vartok[1], "maintainers", strlen(vartok)-2)) {
                    configsection = CONF_MAINTAINERS_SECTION;
               } else {
                    configsection = CONF_REP_SECTION;

                    newconfigtag = calloc(1, sizeof(struct configTag));
                    if (!newconfigtag) {
                          fprintf(stderr, "The system is out of memory\n");
                          return NULL;
                    }

                    newconfigtag->tag =
                        (char *) strndup(&vartok[1], strlen(vartok) - 2);
		    curraltrep = 0;
		    for (i = 0; i<ARCHS_MAX; i++) {
		       newconfigtag->arch[i] = configdefaults.arch[i];
		    }
                    newconfigtag->configdefaults = &configdefaults;

                    if (configdefaults.html_basedir) {
                         strcpy(buf, configdefaults.html_basedir);
                         strncat(buf, newconfigtag->tag, sizeof(buf) - strlen(buf));
                         strncat(buf, "/", sizeof(buf) - strlen(buf));
                         newconfigtag->html_dir = (char *) strdup(buf);
                    }

                    if (!currconfigtag) {
                         firstconfigtag = newconfigtag;
                    } else {
		         currconfigtag->next = newconfigtag;
		    }
                    currconfigtag = newconfigtag;
               }
          } else if (configsection == CONF_REP_SECTION) {
               /* we are reading inside a repository section... */
               if (!strcmp(vartok, "REPOSITORY_DIR")) {
                    currconfigtag->repository_dir =
                         (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "PARENT")) {
                    currconfigtag->repository[curraltrep]=findRepositoryByTag(valuetok);
                    if (!currconfigtag->repository[curraltrep]) {
                        fprintf(stderr,"Error: repository %s requested in %s was not previously configured; aborting.\n",
                           valuetok, currconfigtag->tag);
                        exit(1);
                    }
                    curraltrep++;
               } else if (!strcmp(vartok, "HTML_DIR")) {
                    currconfigtag->html_dir = (char *) strdup(valuetok);
							 } else if (!strcmp(vartok, "DOWNLOAD_PREFIX")) {
                    currconfigtag->download_prefix = (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "DOWNLOAD_DIR")) {
                    currconfigtag->download_dir = (char *) strdup(valuetok);
							 } else if (!strcmp(vartok, "REPODATA_URL")) {
	                  currconfigtag->repodata_url = (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "SHOWFILE_PREFIX")) {
                    currconfigtag->showfile_prefix = (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "REPOSITORY_SOURCE_DIR")) {
                    currconfigtag->repository_source_dir =
                        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "DESCRIPTION")) {
                    currconfigtag->description =
                        (char *) strdup(valuetok);
	       } else if (!strcmp(vartok, "ARCHS")) {
	            vartok = (char *) strtok(valuetok, " ");
	            i = 0;
	            while ((vartok) && (i < ARCHS_MAX)) {
		       currconfigtag->arch[i] = malloc(sizeof vartok);
                       if (!currconfigtag->arch[i]) {
                          fprintf(stderr, "The system is out of memory\n");
                          exit(1);
                       }
                       strcpy(currconfigtag->arch[i], vartok);
		       vartok = strtok(NULL, " ");
		       i++;
		    }
	            if (vartok) {
                       fprintf(stderr,
                            "Error: exceeding number of archs defined; maximum is %d.",ARCHS_MAX);
                            return NULL;
                    } else {
		       currconfigtag->arch[i] = NULL;
		    }
               } else {
                    fprintf(stderr, "Undefined token: %s\n", vartok);
                    return NULL;
               }
          } else if (configsection == CONF_DEFAULTS_SECTION) {
	       /* we are in the default section */
               if (!strcmp(vartok, "HTML_BASEDIR")) {
                    configdefaults.html_basedir =
                        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "DISTRIBUTION_NAME")) {
                    configdefaults.distribution_name =
                        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "URL_ADDRESS")) {
                    configdefaults.url_address =
		        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "URL_PREFIX")) {
                    configdefaults.url_prefix =
		        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "URL_SEARCH_PREFIX")) {
                    configdefaults.url_search_prefix =
		        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "URL_DIR")) {
                    configdefaults.url_dir =
		        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "FAVICON_SEARCH_PREFIX")) {
                    configdefaults.favicon_search_prefix =
                        (char *) strdup(valuetok);
               } else if (!strcmp(vartok, "ARCHS")) {
	            vartok = (char *) strtok(valuetok, " ");
	            i = 0;
	            while ((vartok) && (i < ARCHS_MAX)) {
		       configdefaults.arch[i] = malloc(sizeof vartok);
                       if (!configdefaults.arch[i]) {
                          fprintf(stderr, "The system is out of memory\n");
                          exit(1);
                       }
                       strcpy(configdefaults.arch[i], vartok);
		       vartok = strtok(NULL, " ");
		       i++;
		    }
	            if (vartok) {
                       fprintf(stderr,
                            "Error: exceeding number of archs defined; maximum is %d.",ARCHS_MAX);
                            return NULL;
                    }
               } else {
                    fprintf(stderr,
                            "Invalid token for `defaults' tag: %s\n",
                            vartok);
                    return NULL;
               }
          } else if (configsection == CONF_MAINTAINERS_SECTION) {

              /* find or create maintainer */
              currmaintainer = getPackagerByName(vartok, 1);

              if (!currmaintainer) {
                    fprintf(stderr,
                            "Error: cannot create maintainers; aborting.");
                    return NULL;
	      }

	      currmaintainer->role |= PACKAGER_ROLE_MAINTAINER;

              vartok = (char *) strtok(valuetok, "\"");

	      i = 0;
	      while ((currmaintainer->alias[i]) && (i < PACKAGER_MAXALIASES)) i++;

	      while ((vartok) && (i < PACKAGER_MAXALIASES)) {
	          currmaintainer->alias[i]=strdup(vartok);
		  vartok = strtok(NULL, "\"");
		  vartok = strtok(NULL, "\"");
		  i++;
	      }
	      if (vartok) {
                    fprintf(stderr,
                            "Error: exceeding number of aliases defined for maintainer %s.",currmaintainer->name);
                    return NULL;
	      }
          }
     }
     /* check for errors and complete previous tag infos */
     for(currconfigtag = firstconfigtag; currconfigtag;
         currconfigtag = currconfigtag->next) {

          if (!currconfigtag->repository_dir) {
               fprintf(stderr,
                       "REPOSITORY_DIR not given for tag %s\n",
                       currconfigtag->tag);
               return NULL;
          }

          if (!currconfigtag->repository_source_dir) {
               strcpy(buf, currconfigtag->repository_dir);
               strncat(buf, "/SRPMS.base/", sizeof(buf) - strlen(buf));
               currconfigtag->repository_source_dir =
                    (char *) strdup(buf);
          }

          currconfigtag->repository_level=0;

          for (i = 0; i < ALT_REPS_MAX; i++) {
	     if (currconfigtag->repository[i]) {
	        currconfigtag->repository_level=i+1;
	     }
          }
          currconfigtag->repository[currconfigtag->repository_level]=currconfigtag;

	  if (!currconfigtag->description)
	     currconfigtag->description=currconfigtag->tag;

     }

     fclose(fin);

     return firstconfigtag;
}

char **
dupnargv(char **argv, int count)
{
     int argc;
     char **copy;

     if (argv == NULL)
          return NULL;

     /* the vector */
     copy = (char **) malloc((count + 1) * sizeof(char *));
     if (copy == NULL)
          return NULL;

     /* the strings */
     for (argc = 0; argc < count; argc++) {
          int len = strlen(argv[argc]);

          copy[argc] = malloc(sizeof(char *) * (len + 1));
          if (copy[argc] == NULL) {
               freeargv(copy);
               return NULL;
          }

          strcpy(copy[argc], argv[argc]);
     }

     copy[argc] = NULL;
     return copy;
}

void *
memndup(void *memp, size_t size)
{
     void *newp;
     newp = malloc(size);
     if (!newp)
          return 0;

     memcpy(newp, memp, size);
     return newp;
}

void log_debug_set(int value) {
   debug_log = value;
}

void logmsg(int level, const char *msg, ...) {

   va_list ap;
   static char oldmsg[20][PATH_MAX];
   static int curroldmsg = 0;
   char newmsg[PATH_MAX];
   int i;

   va_start(ap, msg);
   vsnprintf((char*)&newmsg, PATH_MAX, msg, ap);
   va_end(ap);

   for (i = 0; i < 20; i++) {
      /* only log the same message once */
      if (!strncmp(newmsg,(char*)&(oldmsg[i]),PATH_MAX)) return;
    }

   switch (level) {
      case LOG_MARK:
          fprintf(stderr,"Info: ");
		  for (int j=0; j<strlen(newmsg);j++) fprintf(stderr, "*");
          fprintf(stderr,"\nInfo: %s\nInfo: ",(char*)&newmsg);
		  for (int j=0; j<strlen(newmsg);j++) fprintf(stderr, "*");
          fprintf(stderr,"\n");
          break;
      case LOG_WARNING:
          fprintf(stderr,"Warning: %s\n",(char*)&newmsg);
          break;
      case LOG_ERROR:
          fprintf(stderr,"Error: %s\n",(char*)&newmsg);
          break;
      case LOG_DEBUG:
          if (debug_log == 1) fprintf(stderr,"DEBUG: %s\n",(char*)&newmsg);
          break;
      default:
          break;
   }
   if ((level == LOG_WARNING) || (level == LOG_ERROR)) {
      strncpy(oldmsg[curroldmsg++],newmsg,PATH_MAX);
      if (curroldmsg >= 20) curroldmsg = 0;
   }
}

char *
strip_separators(char *st, const char *sepstr)
{
     if (st) {
          int i, beg = -1, end = -1;
          for (i = 0; st[i] != '\0'; i++) {
               if (strchr(sepstr, st[i]) != NULL) {
                    continue;
               }
               else {
                    end = i;
                    if (beg < 0)
                         beg = i;
               }
          }
          if (beg > 0)
               st = memmove(st, (char *) &st[beg], end - beg + 1);
          st[end - beg + 1] = '\0';
     }
     return st;
}

humanDate *
simpleTimeToTemplate(long calendartime, const char* template, humanDate * strdate)
{
     struct tm *ltime;

     ltime = localtime((time_t *) & calendartime);
     strftime((char *) strdate, 16, template, ltime);

     return strdate;
}

humanDate *
simpleTimeToHuman(long calendartime, humanDate * strdate)
{
     simpleTimeToTemplate(calendartime, "%a %h %d %Y", strdate);

     return strdate;
}

void
fprintf_depstable_filtered_var(FILE* f,char* st) {
   int i;
   for (i = 0; st[i] != '\0'; i++) {
      switch (st[i]) {
         case ':': fputc('_',f); break;
         case '(': fputc('[',f); break;
         case ')': fputc(']',f); break;
	 default: fputc(st[i],f); break;
      }
   }
}

char *humanSize(long long sbytes, sizeString *s) {

    if (sbytes < 1024) {
        snprintf((char *)s, SSSIZE, "%lld B", sbytes);
        return (char *)s;
    }
    if ((sbytes / 1024) < 1024) {
        snprintf((char *)s, SSSIZE, "%.2f KB", sbytes / 1024.0);
        return (char*)s;
    }
    if ((sbytes / 1024 / 1024 ) < 1024) {
        snprintf((char *)s, SSSIZE, "%.2f MB", sbytes / 1024.0 / 1024.0);
        return (char*)s;
    }
    if ((sbytes / 1024 / 1024 / 1024 ) < 1024) {
        snprintf((char *)s, SSSIZE, "%.2f GB", sbytes / 1024.0 / 1024.0 / 1024.0);
        return (char*)s;
    }
    snprintf((char *)s, SSSIZE, "%.2f TB", sbytes / 1024.0 / 1024.0 / 1024.0 / 1024.0);
    return (char*)s;
}


const char *htmlclean(const char *source,char *dest,unsigned int max)
{
   unsigned int i=0,j=0;
   char subst[10];

   while (source[i] && i<max) {
      switch (source[i]) {
          case '<': strcpy(subst,"&lt;"); break;
	  case '>': strcpy(subst,"&gt;"); break;
          case '&': strcpy(subst,"&amp;"); break;
	  case '"': strcpy(subst,"&quot;"); break;
	  case '\n': strcpy(subst,"<br>\n"); break;
	  default: subst[0]=source[i]; subst[1]='\0'; break;
      }
      if (j+strlen(subst)<max) {
         strcpy((char*)&dest[j],subst);
	 j+=strlen(subst);
      } else {
         fprintf(stderr,"Warning: can't tidy HTML string due to limited buffer.\n");
         return source;
      }
      i++;
   }
   return dest;
}

const char *htmlcleanNoBr(const char *source,char *dest,unsigned int max)
{
   unsigned int i=0,j=0;
   char subst[10];

   while (source[i] && i<max) {
      switch (source[i]) {
          case '<': strcpy(subst,"&lt;"); break;
	  case '>': strcpy(subst,"&gt;"); break;
          case '&': strcpy(subst,"&amp;"); break;
	  case '"': strcpy(subst,"&quot;"); break;
	  default: subst[0]=source[i]; subst[1]='\0'; break;
      }
      if (j+strlen(subst)<max) {
         strcpy((char*)&dest[j],subst);
	 j+=strlen(subst);
      } else {
         fprintf(stderr,"Warning: can't tidy HTML string due to limited buffer.\n");
	 return source;
      }
      i++;
   }
   return dest;
}

/*
  scansdir is like scandir with the difference that it uses stat()
           to get entry file types

*/
int scansdir(const char *dir, struct dirent ***namelist,
            int (*sel)(const struct dirent *),
            int (*compar)(const struct dirent **, const struct dirent **))
{
  DIR *d;
  char pn[PATH_MAX];
  struct dirent *entry;
  struct stat buf;
  register int i=0;
  int pe;
  size_t entrysize;

  if ((d=opendir(dir)) == NULL)
     return(-1);

  pe=strlen(dir);
  memcpy(pn,dir,pe);

  *namelist=NULL;
  while ((entry=readdir(d)) != NULL)
  {
    if (entry->d_type == DT_UNKNOWN) {
       strncpy(&pn[pe],entry->d_name,PATH_MAX-pe);
       stat(pn,&buf);
       if (S_ISREG(buf.st_mode)) entry->d_type |= DT_REG;
       else if (S_ISDIR(buf.st_mode)) entry->d_type |= DT_DIR;
       else if (S_ISFIFO(buf.st_mode)) entry->d_type |= DT_FIFO;
       else if (S_ISSOCK(buf.st_mode)) entry->d_type |= DT_SOCK;
       else if (S_ISCHR(buf.st_mode)) entry->d_type |= DT_CHR;
       else if (S_ISBLK(buf.st_mode)) entry->d_type |= DT_BLK;
    }

    if (sel == NULL || (sel != NULL && (*sel)(entry)))
    {
      *namelist=(struct dirent **)realloc((void *)(*namelist),
                 (size_t)((i+1)*sizeof(struct dirent *)));
        if (*namelist == NULL) return(-1);
        entrysize=sizeof(struct dirent)-sizeof(entry->d_name)+strlen(entry->d_name)+1;
        (*namelist)[i]=(struct dirent *)malloc(entrysize);
        if ((*namelist)[i] == NULL) return(-1);
        memcpy((*namelist)[i], entry, entrysize);
        i++;
    }
  }
  if (closedir(d)) return(-1);
//  if (i == 0) return(-1);
  if (compar != NULL)
    qsort((void *)(*namelist), (size_t)i, sizeof(struct dirent *),
          (__compar_fn_t) compar);

  return(i);
}

int
get_favicon_from_url(const char* url,char *buf,int bufsize) {

   int i=0,l=0;

   if (configdefaults.favicon_search_prefix != NULL) {
      snprintf(buf, bufsize, "%s%s", configdefaults.favicon_search_prefix, url);
   } else {
      snprintf(buf, bufsize, "%s", url);

      /* expecting urls in the form http://address/... */
      while ((i<bufsize) && buf[i]) {
         if (buf[i] == '/') {
               l++;
               if (l == 3) break;
            }
         i++;
      }
      buf[i] = 0;
      snprintf(&buf[i], bufsize - l, "/favicon.ico");
   }
   return 0;
}
