/*
 * distromatic - tool for RPM based repositories
 *
 * Copyright (C) 2004-2020 by Silvan Calarco <silvan.calarco@mambasoft.it>
 * Copyright (C) 2006 by Davide Madrisan <davide.madrisan@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#if HAVE_STRINGS_H
# include <strings.h>
#endif

#include <stdlib.h>
#include <fcntl.h>
#include <zlib.h>
#include <libiberty/libiberty.h>

#include "functions.h"
#include "rpmfunctions.h"

/*
 * static int getPackageInfo(
 *      Header h, char **name, char **version, char **release,
 *      char **summary); */

struct EVR_s {
    char *data;
    uint32_t e;
    char *v;
    char *r;
} EVR;

void rpmEVRsplit(const char* ver, struct EVR_s *evr) {
    evr->e = 0;
    evr->v = NULL;
    evr->r = NULL;
    evr->data = strdup(ver);
    for (int i = 0; i < strlen(ver); i++) {
        if (evr->data[i] == ':') {
            evr->e = strtoul(ver, NULL, 10);
            evr->data[i] = '\0';
            evr->v = &evr->data[i+1];
        } else if (evr->data[i] == '-') {
            evr->data[i] = '\0';
            evr->r = &evr->data[i+1];
            if (evr->v == NULL) evr->v = evr->data;
        }
    }
    if (evr->v == NULL) evr->v = evr->data;
}

unsigned int checkVersionWithFlags(const char* cmp1, uint_32 flags, const char* cmp2) {
    int vercmp = 0;
    struct EVR_s evr1, evr2;

    rpmEVRsplit(cmp1, &evr1);
    rpmEVRsplit(cmp2, &evr2);

    if (evr1.e == evr2.e) vercmp = 0;
    else if (evr1.e > evr2.e) vercmp = 1;
    else vercmp = -1;

    if (vercmp == 0) vercmp = rpmvercmp(evr1.v, evr2.v);
    if (vercmp == 0 && evr1.r && evr2.r) vercmp = rpmvercmp(evr1.r, evr2.r);

    free(evr1.data);
    free(evr2.data);

    if ((vercmp == 0) && (flags & RPMSENSE_EQUAL)) return 1;
    else if ((vercmp == 1) && (flags & RPMSENSE_LESS)) return 1;
    else if ((vercmp == -1) && (flags & RPMSENSE_GREATER)) return 1;

    return 0;
}

/*
 * get header from a file descriptor of a YUM compressed header file */
int
getHeader(rpmts* ts, char *headerFile, Header * h)
{
     char buffer[HEADERS_BUFFER_SIZE];
     int len;

     /* check if file is a compressed header or a RPM/SRPM */
     if (!strcmp(&headerFile[strlen(headerFile) - 4], ".hdr")) {
          gzFile gzIn;
          gzIn = gzopen(headerFile, "r");

          if (gzIn == NULL) {
               return 1;
          }                     /* can't gzopen file */

          len = gzread(gzIn, buffer, HEADERS_BUFFER_SIZE);
          gzclose(gzIn);

          /* header file is bigger than buffer */
          if (len >= HEADERS_BUFFER_SIZE) {
               return 2;
          }

          *h = headerLoad(buffer);
     } else {
          FD_t fd;

          fd = Fopen(headerFile, "r");

	  if (fd == NULL || Ferror(fd)) {
               return 1;
          }
          int rc;

#if RPM_VERSION >= 0x040100
          rc = rpmReadPackageFile(*ts, fd, headerFile, h);
          Fclose(fd);
#else
          rc = rpmReadPackageHeader(fd, h, NULL, NULL, NULL);
          fdClose(fd);
#endif
#if RPM_VERSION >= 0x040100
          if (rc != RPMRC_OK && rc != RPMRC_NOTTRUSTED && rc != RPMRC_NOKEY) {
#else
          if (rc != 0) {
#endif
               return 1;
          }
     }

     return 0;
}

char *
headerGetStringEntry(Header h, const int tag)
{
    char *st = NULL;

#if RPM_VERSION_MAJOR >= 0x050000
     HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));

     he->tag = tag;
     if (headerGet(h, he, 0) && he->t == RPM_STRING_TYPE && he->c == 1)
        st = strdup(he->p.str);
     else
        st = NULL;
     he->p.ptr = _free(he->p.ptr);
#else
     st = headerGetAsString(h, tag);
#endif
     return st;
}

char **
headerGetStringArrayEntry(Header h, const int tag, int* count)
{
     char **st = NULL;

#if RPM_VERSION_MAJOR >= 0x050000
     HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));

     he->tag = tag;
     if (headerGet(h, he, 0) && he->t == RPM_STRING_ARRAY_TYPE && he->c >= 1) {
        *count = he->c;
        st = dupnargv(he->p.ptr, he->c);
     } else {
        *count = 0;
     }
     he->p.ptr = _free(he->p.ptr);
#else
    rpmtd td = rpmtdNew();
    int i = 0;
    *count = 0;
    if (headerGet(h, tag, td, HEADERGET_MINMEM) && rpmtdType(td) == RPM_STRING_ARRAY_TYPE && rpmtdCount(td) >= 1) {
        *count = rpmtdCount(td);
        if (*count > 0) {
            st = malloc(sizeof(char*) * (*count + 1));
            const char* str;
            rpmtdInit(td);
            while ((str = rpmtdNextString(td))) {
                st[i] = strdup(str);
                i++;
            }
            // Terminator
            st[i] = NULL;
            rpmtdFree(td);
        }
    }
#endif
    return st;
}

long long headerGetUIntEntry(Header h, const int tag) {

    unsigned long long ret = -1;

#if RPM_VERSION_MAJOR >= 0x050000
    HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));

    he->tag = tag;
    if (headerGet(h, he, 0) && he->c == 1) {
        switch (he->t) {
	   case RPM_UINT8_TYPE:
	      ret = *he->p.ui8p;
	      break;
	   case RPM_UINT16_TYPE:
	      ret = *he->p.ui16p;
	      break;
	   case RPM_UINT32_TYPE:
	      ret = *he->p.ui32p;
	      break;
	   case RPM_UINT64_TYPE:
	      ret = *he->p.ui64p;
	      break;
	   default:
	      ret = -1;
	      break;
        }
    }
    he->p.ptr = _free(he->p.ptr);
#else
    ret = headerGetNumber(h, tag);
#endif
    return ret;
}

uint_32 *headerGetUIntArrayEntry(Header h, const int tag, int *count) {
    uint_32 *ret = NULL;

#if RPM_VERSION_MAJOR >= 0x050000
    HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));
    he->tag = tag;
    if (headerGet(h, he, 0) && he->c >= 1) {
        *count = he->c;
      switch (he->t) {
	   case RPM_UINT8_TYPE:
	      ret = memndup(he->p.ui8p, *count);
	      break;
	   case RPM_UINT16_TYPE:
	      ret = memndup(he->p.ui16p, *count * 2);
	      break;
	   case RPM_UINT32_TYPE:
	      ret = memndup(he->p.ui32p, *count * 4);
	      break;
	   case RPM_UINT64_TYPE:
	      ret = memndup(he->p.ui64p, *count * 8);
	      break;
	   default:
	      ret = NULL;
	      break;
        }
    }
    he->p.ptr = _free(he->p.ptr);
#else
    rpmtd td = rpmtdNew();
    int i = 0;
    if (headerGet(h, tag, td, HEADERGET_MINMEM)) {
        *count = rpmtdCount(td);
        ret = malloc(sizeof(uint_32) * *count);
        rpmtdInit(td);
        while (rpmtdNext(td) >= 0) {
            switch (td->type) {
                case RPM_INT8_TYPE:
                    ret[i] = (uint_32)*rpmtdGetChar(td);
                    break;
                case RPM_INT16_TYPE:
                    ret[i] = (uint_32)*rpmtdGetUint16(td);
                    break;
                case RPM_INT32_TYPE:
                    ret[i] = (uint_32)*rpmtdGetUint32(td);
                    break;
                case RPM_INT64_TYPE:
                    fprintf(stderr,"headerGetUIntArrayEntry: uint_64 type converted to uint_32");
                    ret[i] = (uint_32)*rpmtdGetUint64(td);
                    break;
                default:
                    break;
            }
            i++;
        }
        rpmtdFree(td);
    } else {
        *count = 0;
    }
#endif
    return ret;
}

int
getPackageRequires(Header h, char ***requirename, uint_32 **requireflags,
                   char ***requireversion, int *requirecount)
{
     *requirename = headerGetStringArrayEntry(h, RPMTAG_REQUIRENAME, requirecount);
     *requireflags = headerGetUIntArrayEntry(h, RPMTAG_REQUIREFLAGS, requirecount);
     *requireversion = headerGetStringArrayEntry(h, RPMTAG_REQUIREVERSION, requirecount);

     return 0;
}

int
getPackageProvides(Header h, char ***providename, uint_32 **provideflags,
                   char ***provideversion, int *providecount)
{
     *providename = headerGetStringArrayEntry(h, RPMTAG_PROVIDENAME, providecount);
     *provideflags = headerGetUIntArrayEntry(h, RPMTAG_PROVIDEFLAGS, providecount);
     *provideversion = headerGetStringArrayEntry(h, RPMTAG_PROVIDEVERSION, providecount);

     return 0;
}

int
getPackageObsoletes(Header h, char ***obsoletename, uint_32 **obsoleteflags,
                   char ***obsoleteversion, int *obsoletecount)
{
     *obsoletename = headerGetStringArrayEntry(h, RPMTAG_OBSOLETENAME, obsoletecount);
     *obsoleteflags = headerGetUIntArrayEntry(h, RPMTAG_OBSOLETEFLAGS, obsoletecount);
     *obsoleteversion = headerGetStringArrayEntry(h, RPMTAG_OBSOLETEVERSION, obsoletecount);

     return 0;
}

int
getPackageFiles(Header h, uint_32 **dirindexes,
                    char ***dirnames, int *dirnamescount,
                    char ***basenames, int *filenamescount,
                    char ***usernames, char ***groupnames,
                    int_16 **fileflags)
{
     int count;

     /* note: I assume that basenamecount == dirindexescount */
     *dirindexes = headerGetUIntArrayEntry(h, RPMTAG_DIRINDEXES, &count);
     *dirnames = headerGetStringArrayEntry(h, RPMTAG_DIRNAMES, dirnamescount);
     *basenames = headerGetStringArrayEntry(h, RPMTAG_BASENAMES, filenamescount);
     *usernames = headerGetStringArrayEntry(h, RPMTAG_FILEUSERNAME, &count);
     *groupnames = headerGetStringArrayEntry(h, RPMTAG_FILEGROUPNAME, &count);
     *fileflags = headerGetUIntArrayEntry(h, RPMTAG_FILEMODES, &count);

     return 0;
}

/* find name end position assuming a "name-version-release" scheme */
int rpmnameidx(char *filename) {
   int i, f=0, idx=0;
   for (i = strlen(filename)-1; (i > 0) && (f < 2); i--) {
      if (filename[i] == '-') {
         f++;
         idx=i-1;
      }
   }
   return idx;
}

/* compare two rpm filenames by name and version */
int rpmnamecmp(char *filename1, char *filename2, int checkver) {
   int end1=0, end2=0, i;
   char c1,c2;

   /* find idx of base name from filenames */
   end1 = rpmnameidx(filename1);
   end2 = rpmnameidx(filename2);

   /* make comparison, case insensitive */
   for  (i=0; (i <= end1) && (i <= end2); i++) {
      c1=filename1[i];
      c2=filename2[i];
      if ((c1 >= 'A') && (c1 <= 'Z')) c1=c1+'a'-'A';
      if ((c2 >= 'A') && (c2 <= 'Z')) c2=c2+'a'-'A';
      if (c1 < c2) return -1;
      else if (c1 > c2) return 1;
   }

   if (end1 < end2) return -1;
   else if (end1 > end2) return 1;

   /* same name, sort by version */
   if (checkver)
      i=rpmvercmp(&filename2[i+1],&filename1[i+1]) * -1;
   else
      i=0;

   return i;
}

int scanrpmnamecmp(const struct dirent **f1, const struct dirent **f2) {
   return rpmnamecmp((char*)(*f1)->d_name,(char*)(*f2)->d_name,1);
}

char* printrpmversion(char *s, int bufsize, long epoch, char *version, char *release) {

   if (strcmp(release,"")) {
      if (epoch)
         snprintf(s,bufsize,"%ld:%s-%s",epoch,version,release);
      else
         snprintf(s,bufsize,"%s-%s",version,release);
   } else  {
      if (epoch)
         snprintf(s,bufsize,"%ld:%s",epoch,version);
      else
         snprintf(s,bufsize,"%s",version);
   }
   return s;
}

void rpminit() {
#if RPM_VERSION >= 0x040100
   rpmReadConfigFiles(NULL, NULL);
//   ts = rpmtsCreate();
//   rpmtsSetVSFlags(ts, (rpmVSFlags_e)-1);
#endif
}
